//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine) 
// June 2000
//
#ifndef IPPV_H
#define IPPV_H

#include <string>
#include <iostream>
#include <STime.h>
#include "TimeOrderedTable.h"
#include <Dal.h>

// Exception class

class IppvException {
 public:
  IppvException(const string& s) : _label(s) {}
  void print(ostream& os) const {
    os << _label;
  }
 protected:
  string _label;
};

class OutOfRange : public IppvException {
 public:
  OutOfRange(const string& s) : IppvException(s) {}
};

class NoHkRecords : public IppvException {
 public:
  NoHkRecords(const string& s) : IppvException(s) {}
};

class NoTranslationAvailable : public IppvException {
 public:
  NoTranslationAvailable(const string& s) : IppvException(s) {}
};

class ParameterChange : public IppvException {
 public:
  ParameterChange(const string& s) : IppvException(s) {}
};

class NoHousekeepingFile : public IppvException {
 public:
  NoHousekeepingFile(const string& s) : IppvException(s) {}
};

// An IPPV is the value a column in a time-ordered table has between
// two dates.

class Ippv {
 public:
  enum Type { Periodic = 1, NonPeriodic };
  
  virtual ~Ippv() {};
  virtual void fetch(const TimeOrderedTable *tab,
		     const STime& start,
		     const STime& stop) throw(IppvException) = 0;
  virtual void print (ostream& os, const bool verbose = false) const = 0;
  virtual void notAvailable(ostream& os) const = 0;
  virtual const string& name() const = 0;
};

// This is the IPPV factory where we map the IPPV name to a column and
// an optional label.
//
// There are two constructors. One is unconstrained, the other is used
// to specify a selectlib expression that must be applied to the
// time-ordered table before extracting the IPPV. 
//
// Although they are generic, the latter constructor is usually
// required to extract IPPVs in NP HK files, where for a given column
// several rows map onto the same time tag. Then the selectlib
// expression is used to remove this degeneracy.
//
// For instance take the EPIC HBR Buffer Size NPHK:
//
//  TIME HBRID ENDADDR
//   x     1     v1
//   x     2     v2
//   y     1     v3
//   y     2     v4
//
// At TIME=x more than one value of ENDADDR apply. A selection
// expression HBRID=1 removes the ambiguity.
//

class IppvServer {
 public:
  IppvServer();
  ~IppvServer();
  Ippv * ippv(const string& name,
	      const string& colname,
	      const Dal::DataType datatype,
	      const string& label,
	      const Ippv::Type type,
	      const bool& singlevalue) const;

  Ippv * ippv(const string& name,
	      const string& colname,
	      const Dal::DataType datatype,
	      const string& label,
	      const Ippv::Type type,
	      const string& expression,
	      const bool& singlevalue) const;
};

#endif
