//
// OdfObject.cc
//
// Giuseppe Vacanti, European Space Agency and cosine science & computing
// (c) 1998-2001
//
//  The main method is scan(). This finds all correctly named ODF
//  components in directory. Then each component is examined and its
//  main parameters (instrument, date start, date end) extracted and
//  used to build the ODF summary file.
//
//  file_set -- set tallying the names of the dataset that have
//  already been processed. Can be used to prevent processing of some
//  datasets. 
//
//  file_list -- these datasets will be listed in the ODF summary
//  file. Can be used to add files that are not DAL datasets.
//

#include <string>
#include <vector>
#include <map>
#include <set>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <algorithm>

#include <cstdlib>
#include <cassert>

#include <Tracer.h>
#include <metatask.h>
#include <errstr.h>
#include <Dal.h>
#include <param.h>
#include <Xmm.h>
#include <CalServer.h>
#include <Odf.h>
#include <STime.h>
#include <stringutil.h>
#include <SasDalUtils.h>
#include <Directory.h>

#include "ExposureObject.h"
#include "OdfObject.h"
#include "OdfConstituent.h"

#include "IppvExtractor.h"

// Why are these not members?
static unsigned global_expo_counter = 0;
static map<string,unsigned> instrument_expo_counter;

// My full name via the SAS infrastructure.
static string me() {
  PROC RETURN (string(Info::Task::current()->name()) + "(" + string(Info::Task::current()->package()) + "-" + string(Info::Task::current()->version()) + ") [" + string(Info::Task::current()->release()) + "]");
}


static inline string expid(const string& inst, const string& exp){
  return inst + "/" + exp;
}

OdfObject::OdfObject(const string& dir, const bool findinstrumentmodes,
		     const bool usehousekeeping)
  : start_time(STime::latest),
    end_time(STime::earliest),
    _dir(dir),
    _find_instrument_modes(findinstrumentmodes),
    _use_housekeeping(usehousekeeping),
    _valid(false),
    _tpad(90.)
  // 2000-11-13: I have observed up to 90s delay between DATE-OBS
  // and when an HK parameter actually changes value. 
{
  PROC

  // There can be three diffrent ways to call one instrument
  instruments.push_back("M1");
  long_names["M1"] = "EPIC MOS";
  fits_names["M1"] = "EMOS1";

  instruments.push_back("M2");
  long_names["M2"] = "EPIC MOS";
  fits_names["M2"] = "EMOS2";

  instruments.push_back("R1");
  long_names["R1"] = "RGS";
  fits_names["R1"] = "RGS1";

  instruments.push_back("R2");
  long_names["R2"] = "RGS";
  fits_names["R2"] = "RGS2";

  instruments.push_back("PN");
  long_names["PN"] = "EPIC PN";
  fits_names["PN"] = "EPN";

  instruments.push_back("OM");
  long_names["OM"] = "OM"; 
  fits_names["OM"] = "OM";

  // Data modes that contain science data. These are used to determine
  // the camera mode.
  science_data_modes.push_back(Xmm::IMAGING);
  science_data_modes.push_back(Xmm::TIMING);
  science_data_modes.push_back(Xmm::REDUCED_IMAGING);
  science_data_modes.push_back(Xmm::COMPRESSED_TIMING);
  science_data_modes.push_back(Xmm::BURST);
  science_data_modes.push_back(Xmm::DATA_SPECTROSCOPY);
  science_data_modes.push_back(Xmm::HIGH_TIME_RESOLUTION);
  science_data_modes.push_back(Xmm::DATA_FAST);

  // Data modes for which HK parameter extraction is not
  // attempted. This list must not contain any science data mode, but
  // only "auxiliary" data modes for which HK extraction is known to
  // trigger inappropriate warnings. Write why a data mode is being
  // excluded.

  // DISCARDED_LINES_DATA
  // These data files are often sent to the ground after the end of
  // the observation. For this reason they are known to contain the
  // wrong DATE-END. When DATE-END is past the last HK record
  // available, many warnings are issued that only confuse the user. 
  _no_ippv_data_modes.push_back(Xmm::DISCARDED_LINES_DATA);


  scan();

  if(file_list.size() > 0)
    _valid = true;

  ENDPROC
}  

string OdfObject::identifier() const {
  PROC
  RETURN revolution + "_" + proposal + observation_id;
}

// This is the name used by the SAS
string OdfObject::canonicalSummaryName() const {
  PROC
  RETURN revolution + "_" + proposal + "_SCX00000SUM.SAS";
}

// This is the name of the ODS-generated summary file.
string OdfObject::icdosf() const {
  PROC
  RETURN revolution + "_" + proposal + "_SCX00000SUM.ASC";
}

// Use these two methods to manipulate ascii ODF components.
void OdfObject::insertFile(const string name) {
  PROC
  file_list.push_back(name);
  ENDPROC
}

void OdfObject::removeFile(const string name) {
  PROC
  file_list.erase(find(file_list.begin(), file_list.end(), name));
  ENDPROC
}


// Identify all ODF constituents.
void OdfObject::scan() {
  PROC

  bool odfNotKnown = true; // This indicates that we do not yet know
                           // the ODF identifier.

  list<STime> stimes;

  Directory d(_dir);

  // First look only at the *.FIT files flagged as being Scheduled or
  // Unscheduled. These define the observation structure in terms of
  // exposures and active istruments.

  d.filter("????_??????????_??S????????.FIT|????_??????????_??U????????.FIT");
  typedef Directory::const_iterator LI;
  errstr << "Looking for ODF constituents in " + d.name()
	 <<  Sas::message(Msg::App,Msg::Sparse);
  for(LI j = d.begin(); j != d.end(); j++) {
    const DirEntry& e = **j;

    // Assume that *.FIT files are DAL datasets. Put them in the list.

    pair<set<string, less<string> >::const_iterator, bool> p =
      file_set.insert(e.name());
    if(p.second) { // This is a file we have not yet seen
      errstr << "Examining ODF constituent " + e.basename() <<
	Sas::message(Msg::App,Msg::Noisy);

      //FIXME: OdfConstituent should be implemented in terms of OdfSet.
      OdfConstituent oc(e.name());
      OdfSet odfset(e.basename());
      
      if(odfNotKnown){
	odfNotKnown = false;

	// Initialize the observation ID, the revolution number, and
	// the proposal ID. These are unique to one ODF, so we take the
	// first one we find.
      
	observation_id = oc.observation();
	revolution = oc.revolution();
	proposal = oc.proposal();

	errstr << "Identified the following ODF" << endl;
	errstr << "Revolution:  " << revolution << endl;
	errstr << "Observation: " << observation_id << endl;
	errstr << "Proposal:    " << proposal
	       << Sas::message(Msg::App, Msg::Silent);
      } else {

	// From the second constituent on check that it is the very
	// same observation. 

	if(observation_id != oc.observation() ||
	   revolution != oc.revolution() ||
	   proposal != oc.proposal())
	  errstr << "The following observation was identified: " +
	    identifier() + ". But this dataset " + oc.base_name() +
	    " belongs to a different observation." <<
	    Sas::error("MultipleObservations");
      }

      typedef pair<const string, OdfConstituent>  SOC;
      typedef pair<const string, string> SS;

      file_list.push_back(oc.base_name());      
      SOC pf(oc.instrument(),oc);
      file_map_by_inst.insert(pf);      
      SS pm(oc.instrument(),oc.long_exposure());
      exposure_map_by_inst.insert(pm);
                  
      DataSet * set = dataSetServer->open(oc.name(), DataSet::Read,
					  DataSetServer::Low);
      
      // Access by number as some ODF components may not have block
      // names. Also block names encode the block number ...
      
      ExposureObject this_exp;
      this_exp.exposure = oc.exposure();
      this_exp.scheduled = oc.scheduled();

      Block* blk = set->block(0); 
      this_exp.start = STime(blk->attribute("DATE-OBS")->asString());

      // FIXME: EPN discarded lines data sets have the wrong DATE-END
      // This arises because often the data are set to the ground
      // after the observation has been completed. This can also be
      // hours! DATE-START is correct.
      // If the wrong DATE-END was used, the observation would
      // appear to be much longer than it actually is. Moreover, a
      // task that is asking the OAL how long the observation is and
      // then used this info to make use of the attitude data, would
      // discover that for most of the observation there is no valid
      // attitude. Warnings ensue...
      // Solution: Impose a 2*_tpad duration for each exposure.
      if(odfset.dataMode() == Xmm::DISCARDED_LINES_DATA){
	errstr << "Detected a data set with data mode "
	       << odfset.dataMode() << ". This is know to contain the wrong DATE-END. I'm assigning to it an arbitrary duration of "
	       << 2*_tpad
	       << Sas::message(Msg::App, Msg::Silent);
	this_exp.stop = this_exp.start + 2*_tpad;
      }
      else
	this_exp.stop = STime(blk->attribute("DATE-END")->asString());

      errstr << oc.base_name() << " : " << this_exp.start << " - "
	     << this_exp.stop << Sas::message(Msg::App, Msg::Sparse);
      stimes.push_back(this_exp.start);
      stimes.push_back(this_exp.stop);	
      pair<const string, ExposureObject> pe(oc.instrument(), this_exp);

      inst_exp_time_map.insert(pe);

      dataSetServer->close(set);
    } // I have this file already?
  } // loop over all directory entries


  // Count up also the ascii files and the FITS files with X (neither
  // scheduled nor unscheduled) flag. These FITS files are not used to
  // determine the observation duration.
  d.filter("????_??????????_???????????.ASC|????_??????????_??X????????.FIT");
  for(LI j = d.begin(); j != d.end(); j++) {
    const DirEntry &e = **j;
    file_list.push_back(e.basename());      
  }

  if(stimes.size() == 0 || odfNotKnown)
    _valid = false;
  else {
    stimes.sort();
    start_time = *stimes.begin();
    end_time = *(--stimes.end());
  }

 ENDPROC
}

void OdfObject::writeSummary(ostream & os = cout,
			     const bool writepath = false) const {
   PROC

   if(_valid){

     os << "// ----------------------------------" << endl;	
     os << "// XMM-Newton Science Analysis System" << endl;
     os << "// ----------------------------------" << endl;
     os << "//" << endl;
     os << "// ODF Summary File" << endl;
     os << "// By: " << me() << " on " << STime("now") << endl;
     os << "//" << endl;
     if(writepath){
       os << "//" << endl;
       os << "// Directory where the ODF constituents were found. This may have to be edited to match the local file system structure." << endl;
       os << "//" << endl;
       os << "PATH " << _dir << endl;
     }

     observation_rec(os);
     file_details_rec(os);
     
     typedef vector<string>::const_iterator VI;
     for(VI i = instruments.begin(); i != instruments.end(); i++) {
       string inst = *i;
       instrument_details_rec(inst, os);
       typedef multimap<string,string,less<string> >::const_iterator MI;
       MI lb = exposure_map_by_inst.lower_bound(inst);
       MI ub = exposure_map_by_inst.upper_bound(inst);
       set<string, less<string> > exposures;
       for(MI j = lb; j != ub; j++) {
	 pair<set<string, less<string> >::const_iterator, bool>
	   p = exposures.insert((*j).second);
	 if(p.second) {
	   exposure_sub_record(inst, (*j).second, os);
	   configuration_sub_record(inst, (*j).second, os);
	 }
       }
     }
     proposal_summary_rec(os);

   } else {
     errstr << "Could not find a valid ODF in " << _dir
	    << ", hence the summary file will not be written."
	    << Sas::error("NoOdfFound");
   }

   ENDPROC
}

string OdfObject::hkfile(const string& i, const string& type)
{
  // Code the ODF ICD naming convention
  PROC
  RETURN revolution + "_" + proposal + "_" + i +"X" + "00000" + type + "H.FIT";
}  

void OdfObject::observation_rec(ostream & os) const {
  PROC
  errstr << "Writing the observation record." <<
    Sas::message(Msg::App, Msg::Noisy);

  // We are going to update the observation start/end with possibly
  // more accurate data as they are extracted from the HK and/or
  // adjusted in various mannes. All the wrok is done in
  // exposures_start_end_times. 

  vector<STime> t;
  typedef multimap<string,OdfConstituent,less<string> >::const_iterator MI;
  for(unsigned k=0; k<instruments.size(); ++k){
    MI lb = file_map_by_inst.lower_bound(instruments[k]);
    MI ub = file_map_by_inst.upper_bound(instruments[k]);  
    for(MI i = lb; i != ub; i++) {
      vector<STime> v = exposure_start_end_times(instruments[k],
						 (*i).second.long_exposure());
      t.push_back(v[0]);
      t.push_back(v[1]);
    }
  }
  sort(t.begin(), t.end());
  start_time = t[0];
  end_time = t[t.size()-1];
  
  os << "//" << endl;
  os << "// Observation Record" << endl;
  os << "//" << endl;
  os << "OBSERVATION" << endl;
  os << observation_id << " / Observation Identifier" << endl;
  os << revolution << " / Revolution Identifier" << endl;
  os << start_time << " / Observation Start Time" << endl;
  os << end_time << " / Observation End Time" << endl;
//    os << start_time << " / Observation Start Time" << endl;
//    os << end_time << " / Observation End Time" << endl;
  ENDPROC
}

void OdfObject::file_details_rec(ostream & os) const {
  PROC
  errstr << "Writing the file details record." <<
        Sas::message(Msg::App, Msg::Noisy);

    // FIXME: sorting file_list does not compile??!
  vector<string> v(file_list);
  sort(v.begin(), v.end());
  //  sort(file_list.begin(), file_list.end());
  os << "//" << endl;
  os << "// Number of files in the Observation Summary File, and their names" << endl;
  os << "//" << endl;
  os << "FILES" << endl;
  os << file_list.size() << " / Number of Files" << endl;
  typedef vector<string>::const_iterator VI;
  //  for(VI i = file_list.begin(); i != file_list.end(); i++) {
  for(VI i = v.begin(); i != v.end(); i++) {
    const string &s = *i;
    os << s << " / ODF Constituent" << endl;
  }
  ENDPROC
}

bool OdfObject::is_instrument_active(const string& inst) const {
  PROC
  bool a = false;
  if(number_of_exposures_of_instrument(inst) > 0) a = true;
  RETURN a;
}

unsigned int OdfObject::number_of_exposures_of_instrument(const string& inst) const {
  PROC
  // Here equal_range should work, but I cannot make it work.
  typedef multimap<string,OdfConstituent,less<string> >::const_iterator MI;
  MI lb = file_map_by_inst.lower_bound(inst);
  MI ub = file_map_by_inst.upper_bound(inst);  
  set<string,less<string> > s;  
  for(MI i = lb; i != ub; i++) {
    s.insert((*i).second.long_exposure());
  }
  RETURN s.size();
}

void OdfObject::instrument_details_rec(const string& inst,
				       ostream & os) const {
  PROC
  unsigned int n = number_of_exposures_of_instrument(inst);
  string active = "Y";
  if(n == 0) active = "N";    
  os << "//" << endl;
  os << "// Instrument Record" << endl;
  os << "//" << endl;
  os << "INSTRUMENT" << endl;
  os << inst << endl;
  os << active << " / Is this instrument active?" << endl;
  os << n << " / Number of exposures for this instrument" << endl;
  ENDPROC
}

const list<STime> OdfObject::exposure_times(const string& inst,
					    const string& exp) const
{
  PROC

  // FIXME: another contorsion
  // We rely on the caller using the long_exposure.
  assert(exp.size() == 4);
  assert(exp[0] == 'S' || exp[0] == 'U');

  typedef multimap<string,ExposureObject,less<string> >::const_iterator EI;
  EI lb = inst_exp_time_map.lower_bound(inst);
  EI ub = inst_exp_time_map.upper_bound(inst);
  list<STime> stimes;

  //  errstr << "Start/Stop times being considered for " << expid(inst,exp)
  //	 << Sas::message(Msg::App, Msg::Noisy);
  for(EI i = lb; i != ub; i++) {
    const  ExposureObject &e = (*i).second;
    // FIXME: The exposure string is now '?nnn', where ? is one of 'S'
    // 'U', so we have to combine two strings from the ExposureObject,
    // pending a better design.
    if(e.scheduled + e.exposure == exp) {
      stimes.push_back(e.start);
      stimes.push_back(e.stop);	
      //      errstr << e.start << " " << e.stop << Sas::message(Msg::App, Msg::Noisy);
    }
  }
  stimes.sort();

  RETURN stimes;
}

vector<STime> OdfObject::exposure_start_end_times(const string& inst,
						  const string& exp) const
{
  PROC

  list<STime> stimes = exposure_times(inst,exp);

  STime exp_dstart = *stimes.begin();
  STime exp_dend = *(--stimes.end());
  
  // This is OM, and we know that DATE-OBS and DATE-END are not
  // reliable. Try to find the good values in the NPHK.
  if(inst == "OM"){
    Directory d(_dir);
    d.filter("????_??????????_OM??????NPH.FIT");    
    if(d.size() != 1){
      errstr << "Looking for one and only one file matching the pattern "
	     << d.filter() << " (OM non periodic housekeeping) in the directory "
	     << _dir << ". I found " << d.size() << " instead. The
 lack of OM non periodic housekeeping may be a problem later in the
 SAS. Here it makes me use incorrect start/stop times for the observation."
	     << Sas::warning("IncorrectNumberOfDatasets");
    } else {

      DataSet *set = dataSetServer->open((*d.begin())->name(),
					 Dal::Read);
      Table *tab = set->table(0);
      ColumnData *sid = tab->column("SID")->data();
      int16 *v = sid->int16Data();
      TimeOrderedTable *t = new TimeOrderedTable(set->table(0));
      unsigned mark=0;
      for(unsigned k=0; k<t->rows(); ++k){
	//      cout << k << " " << v[k] << endl;
	if(t->time_of_row(k) > exp_dstart){
	  mark = k;
	  //	cout << "Will start looking at row " << mark << endl;
	  break;
	}
      }
      bool have148 = false;
      bool have149 = false;
      for(unsigned k=mark; k<t->rows(); ++k){
	//      cout << k << " " << v[k] << endl;
	if(have148 && have149)
	  break;
	if(v[k] == 148){
	  exp_dstart = t->time_of_row(k);
	  //	cout << "START " << k << " " << v[k] << t.time_of_row(k) << endl;
	  have148 = true;
	}
	if(v[k] == 149){
	  exp_dend = t->time_of_row(k);
	  //	cout << "STOP " << k << " " << v[k] << t.time_of_row(k) << endl;
	  have149 = true;
	}
      }

      delete sid;
      delete t;
      dataSetServer->close(set);
    }
  }

  vector<STime> v(2);
  v[0] = exp_dstart;
  v[1] = exp_dend;

  errstr << expid(inst, exp) << " start = " << exp_dstart
	 << " end = " << exp_dend << Sas::message(Msg::App, Msg::Sparse);

  RETURN v;
}

void OdfObject::exposure_sub_record(const string& inst,
				    const string& exp,
				    ostream & os) const {
  PROC
    errstr << "Writing the exposure sub-record for " << expid(inst,exp)
	   << Sas::message(Msg::App, Msg::Noisy);

  // FIXME: another contorsion
  // We rely on the caller using the long_exposure.
  assert(exp.size() == 4);
  assert(exp[0] == 'S' || exp[0] == 'U');
  bool scheduled = (exp[0] == 'S');

  vector<STime> v = exposure_start_end_times(inst,exp);
  STime exp_dstart = v[0];
  STime exp_dend = v[1];
  
  string proposal_exp_id = proposal + inst + exp;
  string commanded_exp_id = proposal_exp_id;

  os << "//" << endl;
  os << "// Exposure Record" << endl;
  os << "// Observation Level Exposure Index = " << ++global_expo_counter << endl;
  os << "// Instrument " << "[" << inst << "] Level Exposure Identifier = " << exp << endl;
  os << "// Instrument " << "[" << inst << "] Level Exposure Index = " << ++instrument_expo_counter[inst] << endl;
  os << "//" << endl;
  os << "EXPOSURE" << endl;
  os << exp.substr(1,4) << " / Exposure Identifier [also " << exp <<"]" << endl;
  os << (scheduled ? "Y" : "N") << " / Is this a scheduled exposure?" << endl;
  os << "SCIENCE" << " / Exposure Type" << endl;
  os << commanded_exp_id << " / Commanded Exposure Identifier" << endl;
  os << proposal_exp_id << " / Proposal Exposure Identifier" << endl;
  os << exp_dstart << " / Scheduled Start Time" <<endl;
  os << exp_dend << " / Scheduled End Time" << endl;
  os << exp_dstart << " / Actual Start Time" << endl;
  os << exp_dend << " / Actual End Time" << endl;
	    
  ENDPROC
}

void OdfObject::configuration_sub_record(const string& inst,
					 const string& exp,
					 ostream & os) const {
  PROC
    errstr << "Writing the configuration sub-record for "
	   << expid(inst,exp)
	   << Sas::message(Msg::App, Msg::Noisy);

  // FIXME: another contorsion
  // We rely on the caller using the long_exposure.
  assert(exp.size() == 4);
  assert(exp[0] == 'S' || exp[0] == 'U');

  // ccdNr() below returns the CCD number for the X-ray instruments,
  // but the window number for OM. In the comments I always use CCD,
  // but in the messages and in the summary file I use `word'.
  string word = "CCD";
  if(inst == "OM")
    word = "window";
  
  map<unsigned, Xmm::DataMode> dmodes; // map CCD number to CCD data mode
  map<unsigned, Xmm::Mode> cmodes; // map CCD number to CCD calibration mode
  set<Xmm::Mode> ss; // track how many unique calibration modes we have seen
  Xmm::Mode instmode = Xmm::MODE_NONE;

  // The determination of the instrument modes requires access to the
  // CAL, hence the CCF. There is a minimal CCF that will make the CAL
  // happy for this, namely one with only the MODEPARAMS files. This
  // is however not guaranteed to be for ever true. It is mentioned
  // here in case during testing the full CCF is not available.

  bool disable_mode_determination = false;
  if(_find_instrument_modes){
    Directory d(_dir);
    d.filter("????_??????????_" + inst + exp + "?????.FIT");
  
    // If we are here is because we know that there must be some
    // dataset. So it is ok for this to remain an error.
    if(d.size() == 0)
          errstr << "Looking for one or more files matching the pattern "
		 << d.filter() << " in the directory "
		 << _dir << ". I found " << d.size() << " instead. From a previous scan I expected to find at least one file."
		 << Sas::error("IncorrectNumberOfDatasets");

    errstr << "Determination of the data modes for "
	   << expid(inst,exp)
	   << ". Found " << d.size() << " ODF constituents."
	   << Sas::message(Msg::App, Msg::Noisy);
    
    for(Directory::const_iterator k = d.begin(); k != d.end(); ++k){
      OdfSet s((*k)->basename(), _dir);
      errstr << s.basename() << " has data mode " << s.dataMode() <<
	Sas::message(Msg::App, Msg::Noisy);
      
      // We ask about the CCD data mode only if we're dealing with a
      // component that is science related. However, not all
      // science-related ODF components lead to a camera mode (for
      // instance RGS diagnostic images) and that we catch by asking
      // the calServer if it could determine the mode.
      if(s.type() == OdfSet::EVENT_LIST || s.type() == OdfSet::IMAGE){

	// Do I have already a data mode entry for the current CCD? If
	// I do then I expect it to be the same as the one I already have.
	const Xmm::DataMode current_dmode = s.dataMode();

	// The following condition could only happen for EMOS, CCD 1
	// for which a two-node readout mode (leading to two ODF
	// constituents for one CCD) is possible. In this case I
	// expect the CCD data mode to be the same. 

	// WARNING: in at least one case I have seen:
	//
	// 0115_0127720201_PNU00201IME.FIT
	// 0115_0127720201_PNU00201ODI.FIT
	//
	// This triggers the warning.
	//
	// 2000-10-10: there seems to be agreement that such a
	// condition should not happen (ie, it is a bug in the
	// ODS). It seems that its occurrence is related to some
	// manual telecommanding (exposures are unscheduled) that
	// fools the software that should put the data into two
	// different exposures.
	//
	if(dmodes.find(s.ccdNr()) != dmodes.end() &&
	   dmodes[s.ccdNr()] != current_dmode){
	  errstr << "Found inconsistent duplicate " + word + " data mode entry for "
		 << s.basename() << " " + word + " number " << s.ccdNr()
		 << ". The existing entry is: "
		 << dmodes[s.ccdNr()] << endl
		 << "All operations related to the determination of the CCD data mode, calibration mode, and instrument configuration will be disabled for the current exposure."
		 << Sas::warning("InconsistentDataMode");
	  disable_mode_determination = true;
	  break; // no need to continue loping over the ODF constituents
	} else
	  dmodes[s.ccdNr()] = current_dmode;

	// Only CCD data modes that are science data modes (as opposed
	// to science related) are processed to determine the camera
	// mode. 
	vector<Xmm::DataMode>::const_iterator p =
	  find(science_data_modes.begin(), science_data_modes.end(),
	       current_dmode);
	if(p != science_data_modes.end()){
	  Xmm::Mode calibration_mode = Xmm::MODE_NONE;
	  // set the CAL state, but ask the server to be silent about
	  // a failure to determine the mode and the filter.
	  calServer.setState(s.name(), true, true);
	  CalState *cs = calServer.state();
	  ChoiceStateVariable<Mode> *m = cs->mode();
	  if(m->isUncertain()){
	    errstr << "I asked the calibration access layer to determine the data mode for "
		   << s.basename() << " and I did not get a unique answer."
		   << "All operations related to the determination of the CCD data mode, calibration mode, and instrument configuration will be disabled for the current exposure."
		   << Sas::warning("UnknownCalibrationMode");
	    disable_mode_determination = true;
	    break; // no need to continue loping over the ODF constituents	
	  } else { 
	    calibration_mode = m->get();
	    if(cmodes.find(s.ccdNr()) != cmodes.end() &&
	       cmodes[s.ccdNr()] != calibration_mode){
	      // This could only happen for EMOS, CCD 1 for which a two-node
	      // readout mode is possible. Duplicate entries are ok, but the I
	      // expect the CCD camera mode to be the same.
	      //
	      // But see WARNING above.
	      errstr << "Found inconsistent duplicate " + word + " camera mode entry for "
		     << s.basename() << " " + word + " number " << s.ccdNr()
		     << ". The existing entry is: "
		     << cmodes[s.ccdNr()]
		     << "All operations related to the determination of the CCD data mode, calibration mode, and instrument configuration will be disabled for the current exposure."
		     << Sas::warning("InconsistentCalibrationMode");
	      disable_mode_determination = true;
	      break; // no need to continue loping over the ODF constituents
	    } else {
	      cmodes[s.ccdNr()] = calibration_mode;
	      ss.insert(calibration_mode);
	      errstr << s.basename() << " has calibration mode " << calibration_mode
		     << Sas::message(Msg::App, Msg::Verbose);
	    } // good calibration mode
	  } // uncertain calibration mode
	} else {
	  errstr << s.dataMode() << " is not a science data mode."
		 << Sas::message(Msg::App, Msg::Noisy);	  
	} // is a science data mode
      } // is an event list or an image
    } // for each ODF constituent 
  
    if(disable_mode_determination){
      cmodes.clear();
      dmodes.clear();
    }
    if(cmodes.size() != 0){
      // Instrument mode determination:
      //
      //  EMOS: the CAL camera mode for CCD 1 (either node). Expect to
      // see at most 2 data modes (one for CCD 1, one for the other 6).
      
      // RGS and PN: the CAL camera mode for any CCD (they all operate
      // in the same manner). Expect to see only 1 data mode.
      
      // OM: the CAL camera mode for window 0. Expect to see at most 2
      // data modes (IMI with FAE).
      
      // EMOS
      if(inst == "M1" || inst == "M2" || inst == "OM"){
	if(ss.size() > 2){
	  errstr << "While looking at " << expid(inst,exp)
		 << " I found " << ss.size()
		 << " data mode entries, whereas I was expecting no more than 2. Thereby I cannot determine the instrument mode."
		 << Sas::warning("UnknownInstrumentMode");      
	  instmode = Xmm::MODE_NONE;
	} else
	  if(cmodes.find(1) != cmodes.end())
	    instmode = cmodes[1];
	  else {
	    errstr << "While looking at " << expid(inst,exp)
		   << " I asked " + word + " 1 to determine the instrument mode. Apparently there is no data corresponding to " + word + " 1. Thereby I cannot determine the instrument mode." << Sas::warning("UnknownInstrumentMode"); 
	    instmode = Xmm::MODE_NONE;
	  }
      } else { // other than MOS or OM
	if(ss.size() != 1){
	  errstr << "While looking at " << expid(inst,exp)
		 << " I found " << ss.size()
		 << " data mode entries, whereas I was expecting only 1."
		 << Sas::warning("UnknowInstrumentMode");
	  instmode = Xmm::MODE_NONE;
	} else
	  instmode = *ss.begin();
      } // instrument mode determination: MOS vs non MOS
    } // instrument mode determination
  } // was asked to determine the instrument mode

  vector<STime> v = exposure_start_end_times(inst,exp);
  
  STime start = v[0];
  STime stop = v[1];

  IppvExtractorServer ies = IppvExtractorServer();
  IppvExtractor *ie =
    ies.extractor(Xmm::toInstrument(fits_names[inst]), _dir, expid(inst,exp));
  
  os << "//" << endl;
  os << "// Configuration Record for " << inst << endl;
  os << "//" << endl;
  os << "CONFIGURATION " + long_names[inst] << endl;

  unsigned number_of_parameters = 1; // will always write the
                                     // instrument mode, even if set to Xmm::MODE_NONE
  const bool write_modes = _find_instrument_modes &&
    !disable_mode_determination;
  if(write_modes)
    number_of_parameters += cmodes.size() + dmodes.size();

  bool extract_ippv = _use_housekeeping;
  for(map<unsigned,Xmm::DataMode>::const_iterator k = dmodes.begin();
      k != dmodes.end(); ++k){
    if(extract_ippv && find(_no_ippv_data_modes.begin(),
			    _no_ippv_data_modes.end(),
			    (*k).second) != _no_ippv_data_modes.end()){
      extract_ippv = false;
      errstr << "For the data mode " << (*k).second
	     << " no housekeeping data will be extracted." <<
	Sas::message(Msg::App, Msg::Verbose);
    }
  }
    

  if(extract_ippv)
    number_of_parameters += ie->ippvs();

  os << number_of_parameters << " / Number of configuration parameters available"
     << endl;
  os << "MODE = " << Xmm::toString(instmode) << " / Instrument configuration "
     << endl;
  if(write_modes){
    for(map<unsigned,Xmm::Mode>::const_iterator k = cmodes.begin();
	k != cmodes.end(); ++k)
      os << "CALIBRATION_MODE_" << (*k).first << " = " << Xmm::toString((*k).second)
	 << " / Mode used to calibrate events from " + word + " " << (*k).first
	 << endl;
    for(map<unsigned,Xmm::DataMode>::const_iterator k = dmodes.begin();
	k != dmodes.end(); ++k)
      os << "DATA_MODE_" << (*k).first << " = " << Xmm::toString((*k).second)
	 << " / Data mode for " + word + " " << (*k).first
	 << endl;
  }

  if(extract_ippv){
    // Periodic HK packets are sent to the ground every 8s. It is
    // possible that a parameter is changed, the observation started,
    // but the new value is not yet available in the telemetry. 

    // See note about _tpad at the constructor level.

    STime t1(start);
    STime t2(stop);
    if(stop - start > 2*_tpad){
      t1 += _tpad;
      t2 -= _tpad;
    } else 
      errstr << expid(inst,exp) << " is shorter than " << 2*_tpad
	     << "s. The extraction of housekeeping parameters may fail. (In which case warnings will be issued.)"
	     << Sas::message(Msg::App, Msg::Verbose);    
    ie->write(os, t1, t2);
  }  
  ENDPROC
}

void OdfObject::proposal_summary_rec(ostream & os) const {
  PROC
  errstr << "Writing the proposal summary record." <<
    Sas::message(Msg::App, Msg::Noisy);

  Directory d(_dir);
  d.filter(icdosf());
  if(d.size() != 1)
    errstr << "Looking for one and only one file named " << icdosf()
	   << " in the directory " << _dir << ". Instead I found "
	   << d.size() << Sas::error("IncorrectNumberOfDatasets");
  ifstream is((*d.begin())->name().c_str());
  string item = "";
  while(item.substr(0,8) != "PROPOSAL"){    
    is >> item;
  }
  string line;
  getline(is,line); // this eats to the end of line
  os << "//" << endl;
  os << "// Proposal Summary Record [extracted from " << (*d.begin())->basename() << "]" << endl;
  os << "//" << endl;
  os << "PROPOSAL" << endl; 
  while(!is.eof()){
    getline(is,line);
    os << line << endl;;
  }
  is.close();
  ENDPROC
}

const bool& OdfObject::valid() const { PROC RETURN _valid; }
